import json
import os
import re
import glob
import shutil
from contextlib import contextmanager
from functools import partialmethod
from test.support import EnvironmentVarGuard

import pytest
import requests_mock
from click.testing import CliRunner

from sentinelsat import SentinelAPI, InvalidChecksumError
from sentinelsat import SentinelProductsAPI
from sentinelsat.scripts.cli import cli


@pytest.fixture(scope="session")
def run_cli(credentials):
    runner = CliRunner()
    credential_args = ["--user", credentials[0] or "", "--password", credentials[1] or ""]

    @contextmanager
    def nullcontext():
        yield

    def run(*args, **kwargs):
        """Runs the sentinelsat CLI

        Parameters
        ----------
        *args : str
            The command-line arguments
        with_credentials : bool, optional
            Whether the credentials from DHUS_* environment variables are passed to the CLI
            as --user and --password parameters. Defaults to True.
        must_return_nonzero : bool, optiona
            Asserts that the program returns a non-zero (failure) exit code.
            A zero (success) exit code is asserted otherwise. Defaults to false.
        must_raise : Exception, optional
            If an exception type is provided asserts that a exception of this type is raised.
            Implies must_return_nonzero=True.
        **kwargs
            Any remaining keyword arguments are passed directly to CliRunner.invoke().

        Returns
        -------
            The click.testing.Result object returned by CliRunner.invoke().
            Includes an additional products attribute with a list of products printed
            to stdout by the program.
        """
        with_credentials = kwargs.pop("with_credentials", True)
        must_raise = kwargs.pop("must_raise", None)
        must_return_nonzero = kwargs.pop("must_return_nonzero", False) or must_raise is not None

        assert_raises = pytest.raises(must_raise) if must_raise else nullcontext()
        with assert_raises:
            result = runner.invoke(
                cli,
                credential_args + list(args) if with_credentials else args,
                catch_exceptions=must_raise,
                **kwargs
            )
            if must_raise:
                raise result.exception
        if must_return_nonzero:
            assert result.exit_code != 0
        else:
            assert result.exit_code == 0
        result.products = re.findall("^Product .+$", result.output, re.M)
        return result

    return run


@pytest.fixture
def no_auth_environ():
    with EnvironmentVarGuard() as guard:
        # Temporarily unset credential environment variables
        guard.unset("DHUS_USER")
        guard.unset("DHUS_PASSWORD")
        yield


@pytest.fixture
def no_netrc():
    netrcpath = os.path.expanduser("~/.netrc")
    netrcpath_bak = netrcpath + ".bak"
    if os.path.isfile(netrcpath):
        shutil.move(netrcpath, netrcpath_bak)
        try:
            yield
        finally:
            shutil.move(netrcpath_bak, netrcpath)
    else:
        yield


@pytest.fixture
def netrc_from_environ(no_netrc, credentials):
    netrcpath = os.path.expanduser("~/.netrc")
    assert not os.path.exists(netrcpath)
    with open(netrcpath, "w") as f:
        f.write(
            "\n".join(
                [
                    "machine apihub.copernicus.eu",
                    "login {}".format(credentials[0]),
                    "password {}".format(credentials[1]),
                ]
            )
        )
    try:
        yield
    finally:
        os.remove(netrcpath)


@pytest.mark.vcr
@pytest.mark.scihub
def test_cli_gnss(run_cli):
    run_cli(
        "--gnss",
        "-s",
        "20210201",
        "-e",
        "20210202",
        "--producttype",
        "AUX_POEORB",
        "--query",
        "platformserialidentifier=1B",
    )


@pytest.mark.vcr
@pytest.mark.scihub
def test_cli_geometry_alternatives(run_cli, geojson_string, wkt_string):
    run_cli("--geometry", geojson_string, "--end", "20200101", "--limit", "1")
    run_cli("--geometry", wkt_string, "--end", "20200101", "--limit", "1")


@pytest.mark.fast
def test_cli_geometry_WKT_alternative_fail(run_cli):
    result = run_cli(
        "--geometry",
        "POLYGO((-87.27 41.64,-81.56 37.857,-82.617 44.52,-87.2 41.64))",
        "--end",
        "20200101",
        "--limit",
        "1",
        must_return_nonzero=True,
    )
    assert (
        "neither a GeoJSON file with a valid path, a GeoJSON String nor a WKT string."
        in result.output
    )


@pytest.mark.fast
def test_cli_geometry_JSON_alternative_fail(run_cli):
    result = run_cli(
        "--geometry",
        '{"type": "A bad JSON", "features" :[nothing], ([{ ',
        "--end",
        "20200101",
        "--limit",
        "1",
        must_return_nonzero=True,
    )
    assert "geometry string starts with '{' but is not a valid GeoJSON." in result.output


@pytest.mark.fast
def test_no_auth_fail(run_cli, no_netrc, no_auth_environ, geojson_path):
    result = run_cli(
        "--geometry",
        geojson_path,
        "--end",
        "20200101",
        "--limit",
        "1",
        with_credentials=False,
        must_return_nonzero=True,
    )
    assert "--user" in result.output


@pytest.mark.vcr
@pytest.mark.scihub
def test_no_auth_netrc(run_cli, netrc_from_environ, no_auth_environ, geojson_path):
    run_cli(
        "--geometry",
        geojson_path,
        "--end",
        "20200101",
        "--limit",
        "1",
        with_credentials=False,
    )


@pytest.mark.vcr
@pytest.mark.scihub
def test_returned_filesize(run_cli, geojson_path):
    result = run_cli(
        "--geometry", geojson_path, "-s", "20141205", "-e", "20141208", "-q", "producttype=GRD"
    )
    expected = "1 scenes found with a total size of 0.50 GB"
    assert result.output.split("\n")[-2] == expected

    result = run_cli(
        "--geometry", geojson_path, "-s", "20170101", "-e", "20170105", "-q", "producttype=GRD"
    )
    expected = "18 scenes found with a total size of 27.81 GB"
    assert result.output.split("\n")[-2] == expected


@pytest.mark.vcr
@pytest.mark.scihub
def test_cloud_flag_url(run_cli, geojson_path):
    command = [
        "--geometry",
        geojson_path,
        "--url",
        "https://apihub.copernicus.eu/apihub/",
        "-s",
        "20151219",
        "-e",
        "20151220",
        "-c",
        "10",
    ]

    result = run_cli("--sentinel", "2", *command)

    expected = "Product e071bdda-47ec-4434-aafa-00340442bdda - Date: 2015-12-19T14:47:22.029Z, Instrument: MSI, Satellite: Sentinel-2, Size: 403.60 MB"
    assert result.products[0] == expected
    # For order-by test
    assert "0848f6b8-5730-4759-850e-fc9945d42296" not in result.products[1]

    run_cli("--sentinel", "1", *command, must_return_nonzero=True)


@pytest.mark.vcr
@pytest.mark.scihub
def test_order_by_flag(run_cli, geojson_path):
    result = run_cli(
        "--geometry",
        geojson_path,
        "--url",
        "https://apihub.copernicus.eu/apihub/",
        "-s",
        "20151219",
        "-e",
        "20151220",
        "--order-by",
        "platformname,-size",
    )
    # Check that order matches platformname,-size
    sats = []
    sizes = {}
    for prod in result.products:
        m = re.search(r"Satellite: (\S+), Size: (\S+ \S+)", prod)
        assert m, prod
        sats.append(m.group(1))
        sizes.setdefault(m.group(1), []).append(m.group(2))
    assert sats == sorted(sats)
    for sizes_group in sizes.values():
        assert sizes_group == sorted(sizes_group, reverse=True)


@pytest.mark.vcr
@pytest.mark.scihub
def test_sentinel1_flag(run_cli, geojson_path):
    result = run_cli(
        "--geometry",
        geojson_path,
        "--url",
        "https://apihub.copernicus.eu/apihub/",
        "-s",
        "20151219",
        "-e",
        "20151228",
        "--sentinel",
        "1",
    )

    expected = "Product 6a62313b-3d6f-489e-bfab-71ce8d7f57db - Date: 2015-12-24T09:40:34.129Z, Instrument: SAR-C SAR, Mode: VV VH, Satellite: Sentinel-1, Size: 7.7 GB"
    assert expected in result.products


@pytest.mark.vcr
@pytest.mark.scihub
def test_sentinel2_flag(run_cli, geojson_path):
    result = run_cli(
        "--geometry",
        geojson_path,
        "--url",
        "https://apihub.copernicus.eu/apihub/",
        "-s",
        "20151219",
        "-e",
        "20151228",
        "--sentinel",
        "2",
        "--limit",
        "5",
    )
    for prod in result.products:
        assert "Satellite: Sentinel-2" in prod


@pytest.mark.vcr
@pytest.mark.scihub
def test_sentinel3_flag(run_cli, geojson_path):
    result = run_cli(
        "--geometry",
        geojson_path,
        "-s",
        "20161201",
        "-e",
        "20161202",
        "--sentinel",
        "3",
        "--limit",
        "5",
    )
    for prod in result.products:
        assert "Satellite: Sentinel-3" in prod


@pytest.mark.vcr
@pytest.mark.scihub
def test_product_flag(run_cli, geojson_path):
    result = run_cli(
        "--geometry",
        geojson_path,
        "--url",
        "https://apihub.copernicus.eu/apihub/",
        "-s",
        "20161201",
        "-e",
        "20161202",
        "--producttype",
        "SLC",
    )

    expected = "Product 2223103a-3754-473d-9a29-24ef8efa2880 - Date: 2016-12-01T09:30:22.149Z, Instrument: SAR-C SAR, Mode: VV VH, Satellite: Sentinel-1, Size: 7.98 GB"
    assert result.products[3] == expected


@pytest.mark.vcr
@pytest.mark.scihub
def test_instrument_flag(run_cli, geojson_path):
    result = run_cli(
        "--geometry", geojson_path, "-s", "20161201", "-e", "20161202", "--instrument", "SRAL"
    )
    for prod in result.products:
        assert "Instrument: SRAL" in prod
        assert "Date: 2016-12-01" in prod or "Date: 2016-12-02" in prod


@pytest.mark.vcr
@pytest.mark.scihub
def test_limit_flag(run_cli, geojson_path):
    limit = 15
    result = run_cli(
        "--geometry",
        geojson_path,
        "--url",
        "https://apihub.copernicus.eu/apihub/",
        "-s",
        "20161201",
        "-e",
        "20161230",
        "--limit",
        str(limit),
    )

    assert len(result.products) == limit


@pytest.mark.vcr
@pytest.mark.scihub
def test_uuid_search(run_cli):
    uuid = "d8340134-878f-4891-ba4f-4df54f1e3ab4"
    result = run_cli("--uuid", uuid)
    assert len(result.products) == 1
    assert uuid in result.products[0]


@pytest.mark.vcr
@pytest.mark.scihub
def test_name_search(run_cli):
    result = run_cli(
        "--name", "S1A_WV_OCN__2SSV_20150526T211029_20150526T211737_006097_007E78_134A"
    )

    expected = "Product d8340134-878f-4891-ba4f-4df54f1e3ab4 - Date: 2015-05-26T21:10:28.984Z, Instrument: SAR-C SAR, Mode: VV, Satellite: Sentinel-1, Size: 10.65 KB"
    assert result.products[0] == expected


@pytest.mark.vcr
@pytest.mark.scihub
def test_name_search_multiple(run_cli):
    result = run_cli(
        "--name",
        "S1B_IW_GRDH_1SDV_20181007T164414_20181007T164439_013049_0181B7_345E",
        "--name",
        "S1B_IW_GRDH_1SDV_20181007T164349_20181007T164414_013049_0181B7_A8E3",
        "-q",
        "identifier=S1A_WV_OCN__2SSV_20150526T211029_20150526T211737_006097_007E78_134A",
        "-q",
        "identifier=S1A_WV_OCN__2SSV_20150526T211029_20150526T211737_006097_007E78_134A",
        "-q",
        "identifier=S1A_WV_OCN__2SSH_20150603T092625_20150603T093332_006207_008194_521E",
    )

    expected = {
        "Product b2ab53c9-abc4-4481-a9bf-1129f54c9707 - Date: 2018-10-07T16:43:49.773Z, Instrument: SAR-C SAR, Mode: VV VH, Satellite: Sentinel-1, Size: 1.65 GB",
        "Product 9e99eaa6-711e-40c3-aae5-83ea2048949d - Date: 2018-10-07T16:44:14.774Z, Instrument: SAR-C SAR, Mode: VV VH, Satellite: Sentinel-1, Size: 1.65 GB",
        "Product d8340134-878f-4891-ba4f-4df54f1e3ab4 - Date: 2015-05-26T21:10:28.984Z, Instrument: SAR-C SAR, Mode: VV, Satellite: Sentinel-1, Size: 10.65 KB",
        "Product 1f62a176-c980-41dc-b3a1-c735d660c910 - Date: 2015-06-03T09:26:24.921Z, Instrument: SAR-C SAR, Mode: HH, Satellite: Sentinel-1, Size: 10.54 KB",
    }
    assert set(result.products) == expected


@pytest.mark.vcr
@pytest.mark.scihub
def test_repeated_keywords(run_cli):
    uuids = ["d8340134-878f-4891-ba4f-4df54f1e3ab4", "1f62a176-c980-41dc-b3a1-c735d660c910"]
    result = run_cli("-q", "uuid=" + uuids[0], "-q", "uuid=" + uuids[1])
    result_uuids = set(prod.split(" ", 2)[1] for prod in result.products)
    assert result_uuids == set(uuids)


@pytest.mark.vcr
@pytest.mark.scihub
def test_name_search_empty(run_cli):
    run_cli("--name", "", must_raise=ValueError)


@pytest.mark.vcr
@pytest.mark.scihub
def test_footprints_cli(run_cli, tmpdir, geojson_path):
    result = run_cli(
        "--geometry",
        geojson_path,
        "-s",
        "20151219",
        "-e",
        "20151228",
        "--sentinel",
        "2",
        "--footprints",
        str(tmpdir / "test.geojson"),
    )
    assert len(result.products) == 89
    gj_file = tmpdir / "test.geojson"
    assert gj_file.check()
    content = json.loads(gj_file.read_text(encoding="utf-8"))
    assert len(content["features"]) == len(result.products)
    for feature in content["features"]:
        assert len(feature["properties"]) >= 28
        coords = feature["geometry"]["coordinates"]
        assert len(coords[0]) > 3 or len(coords[0][0]) > 3


@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_download_single(run_cli, api, tmpdir, smallest_online_products, monkeypatch):
    # Change default arguments for quicker test.
    # Also, vcrpy is not threadsafe, so only one worker is used.
    monkeypatch.setattr(
        "sentinelsat.SentinelAPI.download_all",
        partialmethod(SentinelAPI.download_all, max_attempts=2, n_concurrent_dl=1),
    )

    product_id = smallest_online_products[0]["id"]
    command = ["--uuid", product_id, "--download", "--path", str(tmpdir)]

    run_cli(*command)

    # The file already exists, should not be re-downloaded
    run_cli(*command)

    # clean up
    for f in tmpdir.listdir():
        f.remove()

    # Prepare a response with an invalid checksum
    url = "https://apihub.copernicus.eu/apihub/odata/v1/Products('%s')?$format=json" % product_id
    json = api.session.get(url).json()
    json["d"]["Checksum"]["Value"] = "00000000000000000000000000000000"

    # Force the download to fail by providing an incorrect checksum
    with requests_mock.mock(real_http=True) as rqst:
        rqst.get(url, json=json)

        # md5 flag set (implicitly), should raise an exception
        run_cli(*command, must_raise=InvalidChecksumError)

    # clean up
    tmpdir.remove()


@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_product_node_download_single(run_cli, api, tmpdir, smallest_online_products, monkeypatch):
    # Change default arguments for quicker test.
    # Also, vcrpy is not threadsafe, so only one worker is used.
    monkeypatch.setattr(
        "sentinelsat.SentinelProductsAPI.download_all",
        partialmethod(SentinelProductsAPI.download_all, max_attempts=2, n_concurrent_dl=1),
    )
    product_id = smallest_online_products[0]["id"]
    command = ["--uuid", product_id, "--download", "--path", str(tmpdir)]

    run_cli(*command)

    # The file already exists, should not be re-downloaded
    run_cli(*command)

    # clean up
    for f in tmpdir.listdir():
        f.remove()

    # Prepare a response with an invalid checksum
    url = "https://apihub.copernicus.eu/apihub/odata/v1/Products('%s')?$format=json" % product_id
    json = api.session.get(url).json()
    json["d"]["Checksum"]["Value"] = "00000000000000000000000000000000"

    # Force the download to fail by providing an incorrect checksum
    with requests_mock.mock(real_http=True) as rqst:
        rqst.get(url, json=json)

        # md5 flag set (implicitly), should raise an exception
        run_cli(*command, must_raise=InvalidChecksumError)

    # clean up
    tmpdir.remove()


@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_product_node_download_single_with_filter(
    run_cli, api, tmpdir, smallest_online_products, monkeypatch
):
    # Change default arguments for quicker test.
    # Also, vcrpy is not threadsafe, so only one worker is used.
    monkeypatch.setattr(
        "sentinelsat.SentinelProductsAPI.download_all",
        partialmethod(SentinelAPI.download_all, max_attempts=2, n_concurrent_dl=1),
    )

    product_id = smallest_online_products[0]["id"]
    command = [
        "--uuid",
        product_id,
        "--download",
        "--path",
        str(tmpdir),
        "--include-pattern",
        "*.kml",
    ]

    run_cli(*command)

    # The file already exists, should not be re-downloaded
    run_cli(*command)

    files = list(glob.glob(str(tmpdir.join("S*.SAFE", "*"))))
    assert len(files) == 2
    basenames = [os.path.basename(filename) for filename in files]
    assert "manifest.safe" in basenames
    assert "preview" in basenames

    files = list(glob.glob(str(tmpdir.join("S*.SAFE", "preview", "*"))))
    assert len(files) == 1
    basenames = [os.path.basename(filename) for filename in files]
    assert "map-overlay.kml" in basenames

    # clean up
    tmpdir.remove()


@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_download_many(run_cli, api, tmpdir, smallest_online_products, monkeypatch):
    # Change default arguments for quicker test.
    # Also, vcrpy is not threadsafe, so only one worker is used.
    monkeypatch.setattr(
        "sentinelsat.SentinelAPI.download_all",
        partialmethod(SentinelAPI.download_all, max_attempts=2, n_concurrent_dl=1),
    )

    ids = sorted(product["id"] for product in smallest_online_products)

    command = ["--download", "--path", str(tmpdir)]
    for id in ids:
        command += ["--uuid", id]

    # Download 3 tiny products
    run_cli(*command)

    # Should not re-download
    run_cli(*command)

    # clean up
    for f in tmpdir.listdir():
        f.remove()

    # Prepare a response with an invalid checksum
    product_id = ids[0]
    url = "https://apihub.copernicus.eu/apihub/odata/v1/Products('%s')?$format=json" % product_id
    json = api.session.get(url).json()
    json["d"]["Checksum"]["Value"] = "00000000000000000000000000000000"

    # Force one download to fail
    with requests_mock.mock(real_http=True) as rqst:
        rqst.get(url, json=json)
        # md5 flag set (implicitly), should raise an exception
        result = run_cli(*command)
        assert "is corrupted" in result.output

    assert tmpdir.join("corrupt_scenes.txt").check()
    with tmpdir.join("corrupt_scenes.txt").open() as f:
        assert product_id in f.read()

    # clean up
    tmpdir.remove()


@pytest.mark.vcr(allow_playback_repeats=True)
@pytest.mark.scihub
def test_download_single_quicklook(run_cli, api, tmpdir, quicklook_products, monkeypatch):
    # Change default arguments for quicker test.
    # Also, vcrpy is not threadsafe, so only one worker is used.
    monkeypatch.setattr(
        "sentinelsat.SentinelAPI.download_all_quicklooks",
        partialmethod(SentinelAPI.download_all_quicklooks, n_concurrent_dl=1),
    )

    id = quicklook_products[0]["id"]
    command = ["--uuid", id, "--quicklook", "--path", str(tmpdir)]

    run_cli(*command)

    # The file already exists, should not be re-downloaded
    run_cli(*command)

    # clean up
    for f in tmpdir.listdir():
        f.remove()

    # Prepare a response with an invalid checksum
    url = "https://apihub.copernicus.eu/apihub/odata/v1/Products('{id}')/Products('Quicklook')/$value".format(
        id=id
    )
    headers = api.session.get(url).headers
    headers["content-type"] = "image/xxxx"

    # Force the download to fail by providing an incorrect content type
    with requests_mock.mock(real_http=True) as rqst:
        rqst.get(url, headers=headers)

        # incorrect content-type, should fail
        result = run_cli(*command)
        assert "Some quicklooks failed: 1" in result.output

    # clean up
    tmpdir.remove()


@pytest.mark.vcr
@pytest.mark.scihub
def test_info_cli(run_cli):
    result = run_cli("--info")
    assert result.output == (
        "HTTPError: 404 Client Error: Not Found for url: https://apihub.copernicus.eu/apihub/api/stub/version\n"
        "Are you trying to get the DHuS version of APIHub?\nTrying again after conversion to DHuS URL\n"
        "DHuS version: 2.4.1\n"
    )


@pytest.mark.vcr
@pytest.mark.scihub
def test_location_cli(run_cli):
    result = run_cli("--location", "Metz", "-s" "20200101", "-e" "20200102", "-l", "1")
    assert "Found" in result.output
    m = re.search(r"Found (\d+) products", result.output)
    assert m, result.output
    assert int(m.group(1)) < 100



"""
Tests methods dealing with converting product info into pandas and geopandas formats.
"""
import sys

import pytest

from sentinelsat import SentinelAPI


@pytest.mark.pandas
@pytest.mark.fast
def test_missing_dependency_dataframe(monkeypatch):
    with pytest.raises(ImportError):
        monkeypatch.setitem(sys.modules, "pandas", None)
        SentinelAPI.to_dataframe({"test": "test"})


@pytest.mark.geopandas
@pytest.mark.fast
def test_missing_dependency_geodataframe(monkeypatch):
    with pytest.raises(ImportError):
        monkeypatch.setitem(sys.modules, "geopandas", None)
        SentinelAPI.to_geodataframe({"test": "tst"})


@pytest.mark.pandas
@pytest.mark.scihub
def test_to_pandas(products):
    df = SentinelAPI.to_dataframe(products)
    assert type(df).__name__ == "DataFrame"
    assert len(products) == len(df)
    assert set(products) == set(df.index)


@pytest.mark.pandas
@pytest.mark.fast
def test_to_pandas_empty():
    df = SentinelAPI.to_dataframe({})
    assert type(df).__name__ == "DataFrame"
    assert len(df) == 0


@pytest.mark.pandas
@pytest.mark.geopandas
@pytest.mark.scihub
def test_to_geopandas(products):
    gdf = SentinelAPI.to_geodataframe(products)
    assert type(gdf).__name__ == "GeoDataFrame"
    print(gdf.unary_union.area)
    assert gdf.unary_union.area == pytest.approx(89.6, abs=0.1)
    assert len(gdf) == len(products)
    assert gdf.crs == "EPSG:4326"


@pytest.mark.pandas
@pytest.mark.geopandas
@pytest.mark.fast
def test_to_geopandas_empty():
    gdf = SentinelAPI.to_geodataframe({})
    assert type(gdf).__name__ == "GeoDataFrame"
    assert len(gdf) == 0



"""
Tests for functionality related to the OData API of SciHub (https://apihub.copernicus.eu/apihub/odata/v1/...)
"""
from datetime import datetime

import pytest
import requests_mock

from sentinelsat import SentinelAPI
from sentinelsat.exceptions import InvalidKeyError, ServerError
from sentinelsat.sentinel import _parse_odata_timestamp
from .conftest import chain, scrub_string


@pytest.mark.fast
def test_convert_timestamp():
    assert _parse_odata_timestamp("/Date(1445588544652)/") == datetime(
        2015, 10, 23, 8, 22, 24, 652000
    )


@pytest.mark.vcr
@pytest.mark.scihub
def test_get_product_odata_short(api, smallest_online_products, read_yaml):
    responses = {}
    for prod in smallest_online_products:
        id = prod["id"]
        responses[id] = api.get_product_odata(id)
    expected = read_yaml("odata_response_short.yml", responses)
    assert sorted(responses) == sorted(expected)


@pytest.mark.scihub
def test_get_product_odata_short_with_missing_online_key(api, vcr):
    uuid = "8df46c9e-a20c-43db-a19a-4240c2ed3b8b"
    expected_short = {
        "id": "8df46c9e-a20c-43db-a19a-4240c2ed3b8b",
        "size": 143549851,
        "md5": "D5E4DF5C38C6E97BF7E7BD540AB21C05",
        "url": "https://apihub.copernicus.eu/apihub/odata/v1/Products('8df46c9e-a20c-43db-a19a-4240c2ed3b8b')/$value",
        "quicklook_url": "https://apihub.copernicus.eu/apihub/odata/v1/Products('8df46c9e-a20c-43db-a19a-4240c2ed3b8b')/Products('Quicklook')/$value",
        "date": datetime(2015, 11, 21, 10, 3, 56, 675000),
        "footprint": "POLYGON((-63.852531 -5.880887,-67.495872 -5.075419,-67.066071 -3.084356,-63.430576 -3.880541,"
        "-63.852531 -5.880887))",
        "title": "S1A_EW_GRDM_1SDV_20151121T100356_20151121T100429_008701_00C622_A0EC",
        "Online": True,
        "Creation Date": datetime(2015, 11, 21, 13, 22, 1, 652000),
        "Ingestion Date": datetime(2015, 11, 21, 13, 22, 4, 992000),
    }

    for _ in range(2):
        # scrub 'Online' key from response
        with vcr.use_cassette(
            "test_get_product_odata_short_with_missing_online_key",
            before_record_response=chain(
                vcr.before_record_response, scrub_string(b'"Online":false,', b"")
            ),
        ):
            response = api.get_product_odata(uuid)
    assert response == expected_short


@pytest.mark.vcr
@pytest.mark.scihub
def test_get_product_odata_full(api, smallest_online_products, read_yaml):
    responses = {}
    for prod in smallest_online_products:
        id = prod["id"]
        responses[id] = api.get_product_odata(id, full=True)
    expected = read_yaml("odata_response_full.yml", responses)
    assert sorted(responses) == sorted(expected)


@pytest.mark.vcr
@pytest.mark.scihub
def test_get_product_info_bad_key(api):
    with pytest.raises(InvalidKeyError) as excinfo:
        api.get_product_odata("invalid-xyz")
    assert str(excinfo.value) == "Invalid key (invalid-xyz) to access Products"


@pytest.mark.mock_api
def test_get_product_odata_scihub_down(read_fixture_file):
    api = SentinelAPI("mock_user", "mock_password")

    request_url = "https://apihub.copernicus.eu/apihub/odata/v1/Products('8df46c9e-a20c-43db-a19a-4240c2ed3b8b')?$format=json"

    with requests_mock.mock() as rqst:
        rqst.get(request_url, text="Mock SciHub is Down", status_code=503)
        with pytest.raises(ServerError) as excinfo:
            api.get_product_odata("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")
        assert str(excinfo.value) == "HTTP status 503: Mock SciHub is Down"

        rqst.get(
            request_url,
            text='{"error":{"code":null,"message":{"lang":"en","value":'
            "\"No Products found with key '8df46c9e-a20c-43db-a19a-4240c2ed3b8b' \"}}}",
            status_code=500,
        )
        with pytest.raises(ServerError) as excinfo:
            api.get_product_odata("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")
        assert (
            str(excinfo.value)
            == "HTTP status 500: No Products found with key '8df46c9e-a20c-43db-a19a-4240c2ed3b8b' "
        )

        rqst.get(request_url, text="Mock SciHub is Down", status_code=200)
        with pytest.raises(ServerError) as excinfo:
            api.get_product_odata("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")
        assert str(excinfo.value) == "HTTP status 200: Mock SciHub is Down"

        # Test with a real "server under maintenance" response
        rqst.get(request_url, text=read_fixture_file("server_maintenance.html"), status_code=502)
        with pytest.raises(ServerError) as excinfo:
            api.get_product_odata("8df46c9e-a20c-43db-a19a-4240c2ed3b8b")
        assert "The Sentinels Scientific Data Hub will be back soon!" in excinfo.value.msg


@pytest.mark.vcr
@pytest.mark.mock_api
@pytest.mark.scihub
def test_is_online(api):
    uuid = "98ca202b-2155-4181-be88-4358b2cbaaa0"
    invalid_uuid = "98ca202b-2155-4181-be88-xxxxxxxxxxxx"

    request_url = "https://apihub.copernicus.eu/apihub/odata/v1/Products('{}')/Online/$value"

    with requests_mock.mock() as rqst:
        rqst.get(request_url.format(uuid), text="true", status_code=200)
        assert api.is_online(uuid) == True

    with requests_mock.mock() as rqst:
        rqst.get(request_url.format(uuid), text="false", status_code=200)
        assert api.is_online(uuid) == False

    with pytest.raises(InvalidKeyError) as excinfo:
        api.is_online(invalid_uuid)



import re
from io import open

from setuptools import find_packages, setup

# Get the long description from the relevant file
with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

with open("sentinelsat/__init__.py", encoding="utf-8") as f:
    version = re.search(r'__version__\s*=\s*"(\S+)"', f.read()).group(1)

setup(
    name="sentinelsat",
    version=version,
    description="Utility to search and download Copernicus Sentinel satellite images",
    long_description=long_description,
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Utilities",
    ],
    keywords="copernicus, sentinel, esa, satellite, download, GIS",
    author="Kersten Clauss",
    author_email="sentinelsat@krstn.eu",
    url="https://github.com/sentinelsat/sentinelsat",
    license="GPLv3+",
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=open("requirements.txt").read().splitlines(),
    extras_require={
        "dev": [
            "pandas",
            "geopandas",
            "shapely",
            "pytest >= 3.6.3",
            "pytest-vcr",
            "pytest-socket",
            "requests-mock",
            "pyyaml",
            "rstcheck",
            "sphinx >= 1.3",
            "sphinx_rtd_theme",
            "flaky",
        ],
    },
    entry_points="""
    [console_scripts]
    sentinelsat=sentinelsat.scripts.cli:cli
    """,
)



import os.path
import re
from glob import glob
from os.path import dirname, abspath

import pytest
import rstcheck

TESTS_DIR = dirname(abspath(__file__))
PROJECT_ROOT_DIR = dirname(TESTS_DIR)

rst_files = list(glob(os.path.join(PROJECT_ROOT_DIR, "*.rst")))
rst_files += list(glob(os.path.join(PROJECT_ROOT_DIR, "docs", "*.rst")))


@pytest.mark.fast
@pytest.mark.parametrize("rst_file", rst_files)
def test_rst(rst_file):
    with open(rst_file) as input_file:
        contents = input_file.read()

    all_errors = []
    errors = rstcheck.check(contents, report_level=2, ignore={"languages": ["python", "bash"]})
    for line_number, error in errors:
        # report only warnings and higher, ignore Python and Bash pseudocode examples
        if "Title underline too short" in error:
            # These are caused by unicode en dashes and can be ignored
            continue

        # Ignore to text roles provided via Sphinx extensions erroneously marked as unrecognized
        m = re.search('Unknown interpreted text role "([^"]+)"', error)
        if m and m.group(1) in ["program", "paramref"]:
            continue

        m = re.search('Unknown directive type "([^"]+)"', error)
        if m and m.group(1) in ["automodule"]:
            continue

        all_errors.append((line_number, error))

    assert len(all_errors) == 0



"""
Handling of GeoJSON, geometries, WKT, etc. GIS-related functionality.
"""
from datetime import datetime

import geojson
import pytest

from sentinelsat import geojson_to_wkt, read_geojson, SentinelAPI, placename_to_wkt


@pytest.mark.fast
def test_boundaries_latitude_more(fixture_path):
    with pytest.raises(ValueError):
        geojson_to_wkt(read_geojson(fixture_path("map_boundaries_lat.geojson")))


@pytest.mark.fast
def test_boundaries_longitude_less(fixture_path):
    with pytest.raises(ValueError):
        geojson_to_wkt(read_geojson(fixture_path("map_boundaries_lon.geojson")))


@pytest.mark.fast
def test_get_coordinates(fixture_path):
    wkt = (
        "POLYGON((-66.2695 -8.0592,-66.2695 0.7031,"
        "-57.3047 0.7031,-57.3047 -8.0592,-66.2695 -8.0592))"
    )
    assert geojson_to_wkt(read_geojson(fixture_path("map.geojson"))) == wkt
    assert geojson_to_wkt(read_geojson(fixture_path("map_z.geojson"))) == wkt
    assert geojson_to_wkt(read_geojson(fixture_path("map_nested.geojson"))) == wkt


@pytest.mark.vcr
@pytest.mark.scihub
def test_footprints_s1(api, test_wkt, read_fixture_file):
    products = api.query(
        test_wkt, (datetime(2014, 10, 10), datetime(2014, 12, 31)), producttype="GRD"
    )

    footprints = api.to_geojson(products)
    for footprint in footprints["features"]:
        assert not footprint["geometry"].errors()

    expected_footprints = geojson.loads(read_fixture_file("expected_search_footprints_s1.geojson"))
    # to compare unordered lists (JSON objects) they need to be sorted or changed to sets
    assert set(footprints) == set(expected_footprints)


@pytest.mark.scihub
def test_footprints_s2(products, fixture_path):
    footprints = SentinelAPI.to_geojson(products)
    for footprint in footprints["features"]:
        assert not footprint["geometry"].errors()

    with open(fixture_path("expected_search_footprints_s2.geojson")) as geojson_file:
        expected_footprints = geojson.loads(geojson_file.read())
    # to compare unordered lists (JSON objects) they need to be sorted or changed to sets
    assert set(footprints) == set(expected_footprints)


@pytest.mark.vcr
@pytest.mark.scihub
def test_placename_to_wkt_valid():
    place_kwargs = ["florida", "ENVELOPE(-87.634896, -79.974306, 31.000968, 24.396308)"]
    # tests wkt response
    assert placename_to_wkt(place_kwargs[0])[0] == place_kwargs[1]


@pytest.mark.vcr
@pytest.mark.fast
def test_placename_to_wkt_invalid():
    # tests empty bbox exception in response to bad query
    with pytest.raises(ValueError) as e:
        wkt = placename_to_wkt("!@#$%^")
    assert 'Unable to find a matching location for "!@#$%^"' in str(e.value)
